@@ -15,7 +15,8 @@ from .layout import replaced
 from .layout.background import BackgroundLayer
 from .matrix import Matrix
 from .stacking import StackingContext
-from .text.ffi import ffi, harfbuzz, pango, units_from_double, units_to_double
+from .text.ffi import ffi, pango, units_from_double, units_to_double
+from .text.fonts import get_hb_object_data
 from .text.line_break import get_last_word_end
 
 SIDES = ('top', 'right', 'bottom', 'left')
@@ -41,7 +42,7 @@ def darken(color):
     hue, saturation, value = rgb_to_hsv(color.red, color.green, color.blue)
     value /= 1.5
     saturation /= 1.25
-    return hsv_to_rgb(hue, saturation, value) + (color.alpha,)
+    return (*hsv_to_rgb(hue, saturation, value), color.alpha)
 
 
 def lighten(color):
@@ -50,7 +51,7 @@ def lighten(color):
     value = 1 - (1 - value) / 1.5
     if saturation:
         saturation = 1 - (1 - saturation) / 1.25
-    return hsv_to_rgb(hue, saturation, value) + (color.alpha,)
+    return (*hsv_to_rgb(hue, saturation, value), color.alpha)
 
 
 def draw_page(page, stream):
@@ -65,14 +66,6 @@ def draw_page(page, stream):
     draw_stacking_context(stream, stacking_context)
 
 
-def draw_box_background_and_border(stream, page, box):
-    if isinstance(box, boxes.TableBox):
-        draw_table(stream, box)
-    else:
-        draw_background(stream, box.background)
-        draw_border(stream, box)
-
-
 def draw_stacking_context(stream, stacking_context):
     """Draw a ``stacking_context`` on ``stream``."""
     # See https://www.w3.org/TR/CSS2/zindex.html
@@ -121,9 +114,10 @@ def draw_stacking_context(stream, stacking_context):
         # Point 2
         if isinstance(box, (boxes.BlockBox, boxes.MarginBox,
                             boxes.InlineBlockBox, boxes.TableCellBox,
-                            boxes.FlexContainerBox)):
+                            boxes.FlexContainerBox, boxes.ReplacedBox)):
             # The canvas background was removed by layout_backgrounds
-            draw_box_background_and_border(stream, stacking_context.page, box)
+            draw_background(stream, box.background)
+            draw_border(stream, box)
 
         with stacked(stream):
             # dont clip the PageBox, see #35
@@ -142,8 +136,11 @@ def draw_stacking_context(stream, stacking_context):
 
             # Point 4
             for block in stacking_context.block_level_boxes:
-                draw_box_background_and_border(
-                    stream, stacking_context.page, block)
+                if isinstance(block, boxes.TableBox):
+                    draw_table(stream, block)
+                else:
+                    draw_background(stream, block.background)
+                    draw_border(stream, block)
 
             # Point 5
             for child_context in stacking_context.float_contexts:
@@ -154,10 +151,8 @@ def draw_stacking_context(stream, stacking_context):
                 draw_inline_level(stream, stacking_context.page, box)
 
             # Point 7
-            for block in [box] + stacking_context.blocks_and_cells:
+            for block in (box, *stacking_context.blocks_and_cells):
                 if isinstance(block, boxes.ReplacedBox):
-                    draw_background(stream, block.background)
-                    draw_border(stream, block)
                     draw_replacedbox(stream, block)
                 elif block.children:
                     if block != box:
@@ -183,10 +178,9 @@ def draw_stacking_context(stream, stacking_context):
         if box.style['opacity'] < 1:
             group_id = stream.id
             stream = original_stream
-            stream.push_state()
-            stream.set_alpha(box.style['opacity'], stroke=True, fill=True)
-            stream.draw_x_object(group_id)
-            stream.pop_state()
+            with stacked(stream):
+                stream.set_alpha(box.style['opacity'], stroke=True, fill=True)
+                stream.draw_x_object(group_id)
 
         stream.end_marked_content()
 
@@ -452,6 +446,12 @@ def draw_border(stream, box):
         draw_column_border()
         return
 
+    # If there's a border image, that takes precedence.
+    if box.style['border_image_source'][0] != 'none' and box.border_image is not None:
+        draw_border_image(box, stream)
+        draw_column_border()
+        return
+
     widths = [getattr(box, f'border_{side}_width') for side in SIDES]
 
     # No border, return early.
@@ -466,15 +466,17 @@ def draw_border(stream, box):
 
     # The 4 sides are solid or double, and they have the same color. Oh yeah!
     # We can draw them so easily!
-    if set(styles) in (set(('solid',)), set(('double',))) and (
-            len(set(colors)) == 1):
+    if set(styles) in (set(('solid',)), set(('double',))) and (len(set(colors)) == 1):
         draw_rounded_border(stream, box, styles[0], colors[0])
         draw_column_border()
         return
 
     # We're not smart enough to find a good way to draw the borders :/. We must
-    # draw them side by side.
-    for side, width, color, style in zip(SIDES, widths, colors, styles):
+    # draw them side by side. Order is not specified, but this one seems to be
+    # close to what other browsers do.
+    values = tuple(zip(SIDES, widths, colors, styles))
+    for index in (2, 3, 1, 0):
+        side, width, color, style = values[index]
         if width == 0 or not color:
             continue
         with stacked(stream):
@@ -487,6 +489,217 @@ def draw_border(stream, box):
     draw_column_border()
 
 
+def draw_border_image(box, stream):
+    """Draw ``box`` border image on ``stream``."""
+    # See https://drafts.csswg.org/css-backgrounds-3/#border-images
+    image = box.border_image
+    width, height, ratio = image.get_intrinsic_size(
+        box.style['image_resolution'], box.style['font_size'])
+    intrinsic_width, intrinsic_height = replaced.default_image_sizing(
+        width, height, ratio, specified_width=None, specified_height=None,
+        default_width=box.border_width(), default_height=box.border_height())
+
+    image_slice = box.style['border_image_slice'][:4]
+    should_fill = box.style['border_image_slice'][4]
+
+    def compute_slice_dimension(dimension, intrinsic):
+        if isinstance(dimension, (int, float)):
+            return min(dimension, intrinsic)
+        else:
+            assert dimension.unit == '%'
+            return min(100, dimension.value) / 100 * intrinsic
+
+    slice_top = compute_slice_dimension(image_slice[0], intrinsic_height)
+    slice_right = compute_slice_dimension(image_slice[1], intrinsic_width)
+    slice_bottom = compute_slice_dimension(image_slice[2], intrinsic_height)
+    slice_left = compute_slice_dimension(image_slice[3], intrinsic_width)
+
+    style_repeat_x, style_repeat_y = box.style['border_image_repeat']
+
+    x, y, w, h, tl, tr, br, bl = box.rounded_border_box()
+    px, py, pw, ph, ptl, ptr, pbr, pbl = box.rounded_padding_box()
+    border_left = px - x
+    border_top = py - y
+    border_right = w - pw - border_left
+    border_bottom = h - ph - border_top
+
+    def compute_outset_dimension(dimension, from_border):
+        if dimension.unit is None:
+            return dimension.value * from_border
+        else:
+            assert dimension.unit == 'px'
+            return dimension.value
+
+    outsets = box.style['border_image_outset']
+    outset_top = compute_outset_dimension(outsets[0], border_top)
+    outset_right = compute_outset_dimension(outsets[1], border_right)
+    outset_bottom = compute_outset_dimension(outsets[2], border_bottom)
+    outset_left = compute_outset_dimension(outsets[3], border_left)
+
+    x -= outset_left
+    y -= outset_top
+    w += outset_left + outset_right
+    h += outset_top + outset_bottom
+
+    def compute_width_adjustment(dimension, original, intrinsic,
+                                 area_dimension):
+        if dimension == 'auto':
+            return intrinsic
+        elif isinstance(dimension, (int, float)):
+            return dimension * original
+        elif dimension.unit == '%':
+            return dimension.value / 100 * area_dimension
+        else:
+            assert dimension.unit == 'px'
+            return dimension.value
+
+    # We make adjustments to the border_* variables after handling outsets
+    # because numerical outsets are relative to border-width, not
+    # border-image-width. Also, the border image area that is used
+    # for percentage-based border-image-width values includes any expanded
+    # area due to border-image-outset.
+    widths = box.style['border_image_width']
+    border_top = compute_width_adjustment(
+        widths[0], border_top, slice_top, h)
+    border_right = compute_width_adjustment(
+        widths[1], border_right, slice_right, w)
+    border_bottom = compute_width_adjustment(
+        widths[2], border_bottom, slice_bottom, h)
+    border_left = compute_width_adjustment(
+        widths[3], border_left, slice_left, w)
+
+    def draw_border_image(x, y, width, height, slice_x, slice_y,
+                          slice_width, slice_height,
+                          repeat_x='stretch', repeat_y='stretch',
+                          scale_x=None, scale_y=None):
+        if 0 in (intrinsic_width, width, slice_width):
+            scale_x = 0
+        else:
+            extra_dx = 0
+            if not scale_x:
+                scale_x = (height / slice_height) if height and slice_height else 1
+            if repeat_x == 'repeat':
+                n_repeats_x = ceil(width / slice_width / scale_x)
+            elif repeat_x == 'space':
+                n_repeats_x = floor(width / slice_width / scale_x)
+                # Space is before the first repeat and after the last,
+                # so there's one more space than repeat.
+                extra_dx = (
+                    (width / scale_x - n_repeats_x * slice_width) / (n_repeats_x + 1))
+            elif repeat_x == 'round':
+                n_repeats_x = max(1, round(width / slice_width / scale_x))
+                scale_x = width / (n_repeats_x * slice_width)
+            else:
+                n_repeats_x = 1
+                scale_x = width / slice_width
+
+        if 0 in (intrinsic_height, height, slice_height):
+            scale_y = 0
+        else:
+            extra_dy = 0
+            if not scale_y:
+                scale_y = (width / slice_width) if width and slice_width else 1
+            if repeat_y == 'repeat':
+                n_repeats_y = ceil(height / slice_height / scale_y)
+            elif repeat_y == 'space':
+                n_repeats_y = floor(height / slice_height / scale_y)
+                # Space is before the first repeat and after the last,
+                # so there's one more space than repeat.
+                extra_dy = (
+                    (height / scale_y - n_repeats_y * slice_height) / (n_repeats_y + 1))
+            elif repeat_y == 'round':
+                n_repeats_y = max(1, round(height / slice_height / scale_y))
+                scale_y = height / (n_repeats_y * slice_height)
+            else:
+                n_repeats_y = 1
+                scale_y = height / slice_height
+
+        if 0 in (scale_x, scale_y):
+            return scale_x, scale_y
+
+        rendered_width = intrinsic_width * scale_x
+        rendered_height = intrinsic_height * scale_y
+        offset_x = rendered_width * slice_x / intrinsic_width
+        offset_y = rendered_height * slice_y / intrinsic_height
+
+        with stacked(stream):
+            stream.rectangle(x, y, width, height)
+            stream.clip()
+            stream.end()
+            stream.transform(e=x - offset_x + extra_dx, f=y - offset_y + extra_dy)
+            stream.transform(a=scale_x, d=scale_y)
+            for i in range(n_repeats_x):
+                for j in range(n_repeats_y):
+                    with stacked(stream):
+                        translate_x = i * (slice_width + extra_dx)
+                        translate_y = j * (slice_height + extra_dy)
+                        stream.transform(e=translate_x, f=translate_y)
+                        stream.rectangle(
+                            offset_x / scale_x, offset_y / scale_y,
+                            slice_width, slice_height)
+                        stream.clip()
+                        stream.end()
+                        image.draw(
+                            stream, intrinsic_width, intrinsic_height,
+                            box.style['image_rendering'])
+
+        return scale_x, scale_y
+
+    # Top left.
+    scale_left, scale_top = draw_border_image(
+        x, y, border_left, border_top, 0, 0, slice_left, slice_top)
+    # Top right.
+    draw_border_image(
+        x + w - border_right, y, border_right, border_top,
+        intrinsic_width - slice_right, 0, slice_right, slice_top)
+    # Bottom right.
+    scale_right, scale_bottom = draw_border_image(
+        x + w - border_right, y + h - border_bottom, border_right, border_bottom,
+        intrinsic_width - slice_right, intrinsic_height - slice_bottom,
+        slice_right, slice_bottom)
+    # Bottom left.
+    draw_border_image(
+        x, y + h - border_bottom, border_left, border_bottom,
+        0, intrinsic_height - slice_bottom, slice_left, slice_bottom)
+    if slice_left + slice_right < intrinsic_width:
+        # Top middle.
+        draw_border_image(
+            x + border_left, y, w - border_left - border_right, border_top,
+            slice_left, 0, intrinsic_width - slice_left - slice_right,
+            slice_top, repeat_x=style_repeat_x)
+        # Bottom middle.
+        draw_border_image(
+            x + border_left, y + h - border_bottom,
+            w - border_left - border_right, border_bottom,
+            slice_left, intrinsic_height - slice_bottom,
+            intrinsic_width - slice_left - slice_right, slice_bottom,
+            repeat_x=style_repeat_x)
+    if slice_top + slice_bottom < intrinsic_height:
+        # Right middle.
+        draw_border_image(
+            x + w - border_right, y + border_top,
+            border_right, h - border_top - border_bottom,
+            intrinsic_width - slice_right, slice_top,
+            slice_right, intrinsic_height - slice_top - slice_bottom,
+            repeat_y=style_repeat_y)
+        # Left middle.
+        draw_border_image(
+            x, y + border_top, border_left, h - border_top - border_bottom,
+            0, slice_top, slice_left,
+            intrinsic_height - slice_top - slice_bottom,
+            repeat_y=style_repeat_y)
+    if (should_fill and slice_left + slice_right < intrinsic_width
+            and slice_top + slice_bottom < intrinsic_height):
+        # Fill middle.
+        draw_border_image(
+            x + border_left, y + border_top, w - border_left - border_right,
+            h - border_top - border_bottom, slice_left, slice_top,
+            intrinsic_width - slice_left - slice_right,
+            intrinsic_height - slice_top - slice_bottom,
+            repeat_x=style_repeat_x, repeat_y=style_repeat_y,
+            scale_x=scale_left or scale_right, scale_y=scale_top or scale_bottom)
+
+
 def clip_border_segment(stream, style, width, side, border_box,
                         border_widths=None, radii=None):
     """Clip one segment of box border.
@@ -731,8 +944,9 @@ def draw_line(stream, x1, y1, x2, y2, thickness, style, color, offset=0):
     assert x1 == x2 or y1 == y2  # Only works for vertical or horizontal lines
 
     with stacked(stream):
-        stream.set_color_rgb(*color[:3], stroke=True)
-        stream.set_alpha(color[3], stroke=True)
+        if style not in ('ridge', 'groove'):
+            stream.set_color_rgb(*color[:3], stroke=True)
+            stream.set_alpha(color[3], stroke=True)
 
         if style == 'dashed':
             stream.set_dash([5 * thickness], offset)
@@ -751,6 +965,25 @@ def draw_line(stream, x1, y1, x2, y2, thickness, style, color, offset=0):
                 stream.line_to(x2, y2 - thickness / 3)
                 stream.move_to(x1, y1 + thickness / 3)
                 stream.line_to(x2, y2 + thickness / 3)
+        elif style in ('ridge', 'groove'):
+            stream.set_line_width(thickness / 2)
+            stream.set_color_rgb(*color[0][:3], stroke=True)
+            stream.set_alpha(color[0][3], stroke=True)
+            if x1 == x2:
+                stream.move_to(x1 + thickness / 4, y1)
+                stream.line_to(x2 + thickness / 4, y2)
+            elif y1 == y2:
+                stream.move_to(x1, y1 + thickness / 4)
+                stream.line_to(x2, y2 + thickness / 4)
+            stream.stroke()
+            stream.set_color_rgb(*color[1][:3], stroke=True)
+            stream.set_alpha(color[1][3], stroke=True)
+            if x1 == x2:
+                stream.move_to(x1 - thickness / 4, y1)
+                stream.line_to(x2 - thickness / 4, y2)
+            elif y1 == y2:
+                stream.move_to(x1, y1 - thickness / 4)
+                stream.line_to(x2, y2 - thickness / 4)
         elif style == 'wavy':
             assert y1 == y2  # Only allowed for text decoration
             up = 1
@@ -791,10 +1024,9 @@ def draw_outlines(stream, box):
                     stream, outline_box, 4 * (width,), style,
                     styled_color(style, color, side))
 
-    if isinstance(box, boxes.ParentBox):
-        for child in box.children:
-            if isinstance(child, boxes.Box):
-                draw_outlines(stream, child)
+    for child in box.children:
+        if isinstance(child, boxes.Box):
+            draw_outlines(stream, child)
 
 
 def draw_table(stream, table):
@@ -966,24 +1198,21 @@ def draw_replacedbox(stream, box):
         return
 
     with stacked(stream):
-        rounded_box_path(stream, box.rounded_content_box())
-        stream.clip()
-        stream.end()
+        stream.set_alpha(1)
         stream.transform(e=draw_x, f=draw_y)
-        stream.push_state()
-        # TODO: Use the real intrinsic size here, not affected by
-        # 'image-resolution'?
-        box.replacement.draw(
-            stream, draw_width, draw_height, box.style['image_rendering'])
-        stream.pop_state()
+        with stacked(stream):
+            # TODO: Use the real intrinsic size here, not affected by
+            # 'image-resolution'?
+            box.replacement.draw(
+                stream, draw_width, draw_height, box.style['image_rendering'])
 
 
 def draw_inline_level(stream, page, box, offset_x=0, text_overflow='clip',
                       block_ellipsis='none'):
     if isinstance(box, StackingContext):
         stacking_context = box
-        assert isinstance(
-            stacking_context.box, (boxes.InlineBlockBox, boxes.InlineFlexBox))
+        assert isinstance(stacking_context.box, (
+            boxes.InlineBlockBox, boxes.InlineFlexBox, boxes.InlineGridBox))
         draw_stacking_context(stream, stacking_context)
     else:
         draw_background(stream, box.background)
@@ -1033,9 +1262,7 @@ def draw_text(stream, textbox, offset_x, text_overflow, block_ellipsis):
         return
 
     text_decoration_values = textbox.style['text_decoration_line']
-    text_decoration_color = textbox.style['text_decoration_color']
-    if text_decoration_color == 'currentColor':
-        text_decoration_color = textbox.style['color']
+    text_decoration_color = get_color(textbox.style, 'text_decoration_color')
     if 'overline' in text_decoration_values:
         thickness = textbox.pango_layout.underline_thickness
         offset_y = (
@@ -1077,10 +1304,9 @@ def draw_text(stream, textbox, offset_x, text_overflow, block_ellipsis):
 
 def draw_emojis(stream, font_size, x, y, emojis):
     for image, font, a, d, e, f in emojis:
-        stream.push_state()
-        stream.transform(a=a, d=d, e=x + e * font_size, f=y + f)
-        image.draw(stream, font_size, font_size, None)
-        stream.pop_state()
+        with stacked(stream):
+            stream.transform(a=a, d=d, e=x + e * font_size, f=y + f)
+            image.draw(stream, font_size, font_size, None)
 
 
 def draw_first_line(stream, textbox, text_overflow, block_ellipsis, matrix):
@@ -1223,14 +1449,8 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, matrix):
             previous_utf8_position = utf8_position
 
             if font.svg:
-                hb_font = pango.pango_font_get_hb_font(pango_font)
-                hb_face = harfbuzz.hb_font_get_face(hb_font)
-                hb_blob = ffi.gc(
-                    harfbuzz.hb_ot_color_glyph_reference_svg(hb_face, glyph),
-                    harfbuzz.hb_blob_destroy)
-                hb_data = harfbuzz.hb_blob_get_data(hb_blob, stream.length)
-                if hb_data != ffi.NULL:
-                    svg_data = ffi.unpack(hb_data, int(stream.length[0]))
+                svg_data = get_hb_object_data(font.hb_face, 'svg', glyph)
+                if svg_data:
                     # Do as explained in specification
                     # https://learn.microsoft.com/typography/opentype/spec/svg
                     tree = ElementTree.fromstring(svg_data)
@@ -1245,13 +1465,8 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, matrix):
                     a = d = font.widths[glyph] / 1000 / font.upem * font_size
                     emojis.append([image, font, a, d, x_advance, 0])
             elif font.png:
-                hb_font = pango.pango_font_get_hb_font(pango_font)
-                hb_blob = ffi.gc(
-                    harfbuzz.hb_ot_color_glyph_reference_png(hb_font, glyph),
-                    harfbuzz.hb_blob_destroy)
-                hb_data = harfbuzz.hb_blob_get_data(hb_blob, stream.length)
-                if hb_data != ffi.NULL:
-                    png_data = ffi.unpack(hb_data, int(stream.length[0]))
+                png_data = get_hb_object_data(font.hb_font, 'png', glyph)
+                if png_data:
                     pillow_image = Image.open(BytesIO(png_data))
                     image_id = f'{font.hash}{glyph}'
                     image = RasterImage(pillow_image, image_id, png_data)
