@@ -11,6 +11,7 @@ from .. import html
 from ..css import properties, targets
 from ..layout.table import collapse_table_borders
 from ..logger import LOGGER
+from ..text.constants import get_lang_quotes
 from . import boxes
 
 # Maps values of the ``display`` CSS property to box types.
@@ -27,6 +28,9 @@ BOX_TYPE_FROM_DISPLAY = {
     ('block', 'flex'): boxes.FlexBox,
     ('inline', 'flex'): boxes.InlineFlexBox,
 
+    ('block', 'grid'): boxes.GridBox,
+    ('inline', 'grid'): boxes.InlineGridBox,
+
     ('table-row',): boxes.TableRowBox,
     ('table-row-group',): boxes.TableRowGroupBox,
     ('table-header-group',): boxes.TableRowGroupBox,
@@ -50,6 +54,7 @@ def create_anonymous_boxes(box):
     """Create anonymous boxes in box descendants according to layout rules."""
     box = anonymous_table_boxes(box)
     box = flex_boxes(box)
+    box = grid_boxes(box)
     box = inline_in_block(box)
     box = block_in_inline(box)
     return box
@@ -369,7 +374,7 @@ def marker_to_box(element, state, parent_style, style_for, get_image_from_uri,
 def compute_content_list(content_list, parent_box, counter_values, css_token,
                          parse_again, target_collector, counter_style,
                          get_image_from_uri=None, quote_depth=None,
-                         quote_style=None, context=None, page=None,
+                         quote_style=None, lang=None, context=None, page=None,
                          element=None):
     """Compute and return the boxes corresponding to the ``content_list``.
 
@@ -501,11 +506,14 @@ def compute_content_list(content_list, parent_box, counter_values, css_token,
                 break
         elif type_ == 'quote' and None not in (quote_depth, quote_style):
             is_open = 'open' in value
-            insert = not value.startswith('no-')
+            insert = not value.startswith('no-') and quote_style != 'none'
             if not is_open:
                 quote_depth[0] = max(0, quote_depth[0] - 1)
             if insert:
-                open_quotes, close_quotes = quote_style
+                if quote_style == 'auto':
+                    open_quotes, close_quotes = get_lang_quotes(lang)
+                else:
+                    open_quotes, close_quotes = quote_style
                 quotes = open_quotes if is_open else close_quotes
                 add_text(quotes[min(quote_depth[0], len(quotes) - 1)])
             if is_open:
@@ -589,7 +597,7 @@ def content_to_boxes(style, parent_box, quote_depth, counter_values,
     box_list = compute_content_list(
         style['content'], parent_box, counter_values, css_token, parse_again,
         target_collector, counter_style, get_image_from_uri, quote_depth,
-        style['quotes'], context, page)
+        style['quotes'], style['lang'], context, page)
     return box_list or []
 
 
@@ -1034,6 +1042,47 @@ def flex_children(box, children):
         return children
 
 
+def grid_boxes(box):
+    """Remove and add boxes according to the grid model.
+
+    Take and return a ``Box`` object.
+
+    See https://drafts.csswg.org/css-grid-2/#grid-item
+
+    """
+    if not isinstance(box, boxes.ParentBox) or box.is_running():
+        return box
+
+    # Do recursion.
+    children = [grid_boxes(child) for child in box.children]
+    box.children = grid_children(box, children)
+    return box
+
+
+def grid_children(box, children):
+    if isinstance(box, boxes.GridContainerBox):
+        grid_children = []
+        for child in children:
+            if not child.is_absolutely_positioned():
+                child.is_grid_item = True
+            if isinstance(child, boxes.TextBox) and not child.text.strip(' '):
+                # TODO: ignore texts only containing "characters that can be
+                # affected by the white-space property"
+                # https://drafts.csswg.org/css-grid-2/#grid-item
+                continue
+            if isinstance(child, boxes.InlineLevelBox):
+                anonymous = boxes.BlockBox.anonymous_from(child, [child])
+                anonymous.style = child.style
+                child.is_grid_item = False
+                anonymous.is_grid_item = True
+                grid_children.append(anonymous)
+            else:
+                grid_children.append(child)
+        return grid_children
+    else:
+        return children
+
+
 def process_whitespace(box, following_collapsible_space=False):
     """First part of "The 'white-space' processing model".
 
