@@ -21,7 +21,7 @@ class AbsolutePlaceholder:
         object.__setattr__(self, '_layout_done', True)
 
     def translate(self, dx=0, dy=0, ignore_floats=False):
-        if dx == 0 and dy == 0:
+        if dx == dy == 0:
             return
         if self._layout_done:
             self._box.translate(dx, dy, ignore_floats)
@@ -48,7 +48,7 @@ class AbsolutePlaceholder:
 
 @handle_min_max_width
 def absolute_width(box, context, cb_x, cb_y, cb_width, cb_height):
-    # http://www.w3.org/TR/CSS2/visudet.html#abs-replaced-width
+    # https://www.w3.org/TR/CSS2/visudet.html#abs-replaced-width
     ltr = (
         box.style.parent_style is None or
         box.style.parent_style['direction'] == 'ltr')
@@ -122,7 +122,7 @@ def absolute_width(box, context, cb_x, cb_y, cb_width, cb_height):
 
 
 def absolute_height(box, context, cb_x, cb_y, cb_width, cb_height):
-    # http://www.w3.org/TR/CSS2/visudet.html#abs-non-replaced-height
+    # https://www.w3.org/TR/CSS2/visudet.html#abs-non-replaced-height
     paddings_borders = (
         box.padding_top + box.padding_bottom +
         box.border_top_width + box.border_bottom_width)
@@ -178,6 +178,7 @@ def absolute_block(context, box, containing_block, fixed_boxes, bottom_space,
                    skip_stack, cb_x, cb_y, cb_width, cb_height):
     from .block import block_container_layout
     from .flex import flex_layout
+    from .grid import grid_layout
 
     translate_box_width, translate_x = absolute_width(
         box, context, cb_x, cb_y, cb_width, cb_height)
@@ -200,11 +201,16 @@ def absolute_block(context, box, containing_block, fixed_boxes, bottom_space,
             context, box, bottom_space, skip_stack, page_is_empty=True,
             absolute_boxes=absolute_boxes, fixed_boxes=fixed_boxes,
             adjoining_margins=None, discard=False, max_lines=None)
-    else:
+    elif isinstance(box, (boxes.FlexContainerBox)):
         new_box, resume_at, _, _, _ = flex_layout(
             context, box, bottom_space, skip_stack, containing_block,
             page_is_empty=True, absolute_boxes=absolute_boxes,
             fixed_boxes=fixed_boxes)
+    elif isinstance(box, (boxes.GridContainerBox)):
+        new_box, resume_at, _, _, _ = grid_layout(
+            context, box, bottom_space, skip_stack, containing_block,
+            page_is_empty=True, absolute_boxes=absolute_boxes,
+            fixed_boxes=fixed_boxes)
 
     for child_placeholder in absolute_boxes:
         absolute_layout(
@@ -229,13 +235,14 @@ def absolute_layout(context, placeholder, containing_block, fixed_boxes,
         context, box, containing_block, fixed_boxes, bottom_space, skip_stack)
     placeholder.set_laid_out_box(new_box)
     if resume_at:
-        context.broken_out_of_flow.append((box, containing_block, resume_at))
+        context.broken_out_of_flow[placeholder] = (
+            box, containing_block, resume_at)
 
 
 def absolute_box_layout(context, box, containing_block, fixed_boxes,
                         bottom_space, skip_stack):
     # TODO: handle inline boxes (point 10.1.4.1)
-    # http://www.w3.org/TR/CSS2/visudet.html#containing-block-details
+    # https://www.w3.org/TR/CSS2/visudet.html#containing-block-details
     if isinstance(containing_block, boxes.PageBox):
         cb_x = containing_block.content_box_x()
         cb_y = containing_block.content_box_y()
@@ -251,16 +258,15 @@ def absolute_box_layout(context, box, containing_block, fixed_boxes,
     resolve_position_percentages(box, (cb_width, cb_height))
 
     context.create_block_formatting_context()
-    # Absolute tables are wrapped into block boxes
-    if isinstance(box, (boxes.BlockBox, boxes.FlexContainerBox)):
-        new_box, resume_at = absolute_block(
-            context, box, containing_block, fixed_boxes, bottom_space,
-            skip_stack, cb_x, cb_y, cb_width, cb_height)
-    else:
-        assert isinstance(box, boxes.BlockReplacedBox)
+    if isinstance(box, boxes.BlockReplacedBox):
         new_box = absolute_replaced(
             context, box, cb_x, cb_y, cb_width, cb_height)
         resume_at = None
+    else:
+        # Absolute tables are wrapped into block boxes
+        new_box, resume_at = absolute_block(
+            context, box, containing_block, fixed_boxes, bottom_space,
+            skip_stack, cb_x, cb_y, cb_width, cb_height)
     context.finish_block_formatting_context(new_box)
     return new_box, resume_at
 
@@ -271,7 +277,7 @@ def absolute_replaced(context, box, cb_x, cb_y, cb_width, cb_height):
         box.style.parent_style is None or
         box.style.parent_style['direction'] == 'ltr')
 
-    # http://www.w3.org/TR/CSS21/visudet.html#abs-replaced-width
+    # https://www.w3.org/TR/CSS21/visudet.html#abs-replaced-width
     if box.left == box.right == 'auto':
         # static position:
         if ltr:
@@ -307,7 +313,7 @@ def absolute_replaced(context, box, cb_x, cb_y, cb_width, cb_height):
         else:
             box.left = cb_width - (box.margin_width() + box.right)
 
-    # http://www.w3.org/TR/CSS21/visudet.html#abs-replaced-height
+    # https://www.w3.org/TR/CSS21/visudet.html#abs-replaced-height
     if box.top == box.bottom == 'auto':
         box.top = box.position_y - cb_y
     if 'auto' in (box.top, box.bottom):
@@ -317,9 +323,9 @@ def absolute_replaced(context, box, cb_x, cb_y, cb_width, cb_height):
             box.margin_bottom = 0
         remaining = cb_height - box.margin_height()
         if box.top == 'auto':
-            box.top = remaining
+            box.top = remaining - box.bottom
         if box.bottom == 'auto':
-            box.bottom = remaining
+            box.bottom = remaining - box.top
     elif 'auto' in (box.margin_top, box.margin_bottom):
         remaining = cb_height - (box.border_height() + box.top + box.bottom)
         if box.margin_top == box.margin_bottom == 'auto':
