@@ -7,6 +7,7 @@ from .absolute import AbsolutePlaceholder, absolute_layout
 from .column import columns_layout
 from .flex import flex_layout
 from .float import avoid_collisions, float_layout, get_clearance
+from .grid import grid_layout
 from .inline import iter_line_boxes
 from .min_max import handle_min_max_width
 from .percent import resolve_percentages, resolve_position_percentages
@@ -48,8 +49,7 @@ def block_level_layout(context, box, bottom_space, skip_stack,
                     if not context.forced_break:
                         box.margin_top = 0
 
-        collapsed_margin = collapse_margin(
-            adjoining_margins + [box.margin_top])
+        collapsed_margin = collapse_margin([*adjoining_margins, box.margin_top])
         box.clearance = get_clearance(context, box, collapsed_margin)
         if box.clearance is not None:
             top_border_edge = box.position_y + collapsed_margin + box.clearance
@@ -82,9 +82,13 @@ def block_level_layout_switch(context, box, bottom_space, skip_stack,
         result = flex_layout(
             context, box, bottom_space, skip_stack, containing_block,
             page_is_empty, absolute_boxes, fixed_boxes)
+    elif isinstance(box, boxes.GridBox):
+        result = grid_layout(
+            context, box, bottom_space, skip_stack, containing_block,
+            page_is_empty, absolute_boxes, fixed_boxes)
     else:  # pragma: no cover
         raise TypeError(f'Layout for {type(box).__name__} not handled yet')
-    return result + (None,)
+    return (*result, None)
 
 
 def block_box_layout(context, box, bottom_space, skip_stack,
@@ -111,7 +115,7 @@ def block_box_layout(context, box, bottom_space, skip_stack,
                     context, box, bottom_space, skip_stack,
                     containing_block, page_is_empty, absolute_boxes,
                     fixed_boxes, adjoining_margins)
-        return result + (None,)
+        return (*result, None)
     elif box.is_table_wrapper:
         table_wrapper_width(
             context, box, (containing_block.width, containing_block.height))
@@ -469,7 +473,7 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
                     if not context.forced_break:
                         child_margin_top = 0
             new_collapsed_margin = collapse_margin(
-                adjoining_margins + [child_margin_top])
+                [*adjoining_margins, child_margin_top])
             collapsed_margin_difference = (
                 new_collapsed_margin - old_collapsed_margin)
             for previous_new_child in new_children:
@@ -645,7 +649,8 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     collapsing_with_children = not (
         box.border_top_width or box.padding_top or box.is_flex_item or
-        establishes_formatting_context(box) or box.is_for_root_element)
+        box.is_grid_item or establishes_formatting_context(box) or
+        box.is_for_root_element)
     if collapsing_with_children:
         # Not counting margins in adjoining_margins, if any
         # (there are not padding or borders, see above)
@@ -742,7 +747,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                 max_lines)
         elif stop:
             if box.height != 'auto':
-                if position_y > box.position_y + box.height:
+                if context.overflows(box.position_y + box.height, position_y):
                     # Box heigh is fixed and it doesn’t overflow page, forget
                     # overflowing children.
                     resume_at = None
@@ -1013,7 +1018,7 @@ def find_earlier_page_break(context, children, absolute_boxes, fixed_boxes):
                 if result:
                     new_grand_children, resume_at = result
                     new_child = child.copy_with_children(new_grand_children)
-                    new_children = list(children[:index]) + [new_child]
+                    new_children = [*children[:index], new_child]
 
                     # Re-add footer at the end of split table
                     # TODO: fix table height and footer position
