@@ -10,7 +10,7 @@ def columns_layout(context, box, bottom_space, skip_stack, containing_block,
                    page_is_empty, absolute_boxes, fixed_boxes,
                    adjoining_margins):
     """Lay out a multi-column ``box``."""
-    from .block import (
+    from .block import (  # isort:skip
         block_box_layout, block_level_layout, block_level_width,
         collapse_margin, remove_placeholders)
 
@@ -26,6 +26,10 @@ def columns_layout(context, box, bottom_space, skip_stack, containing_block,
         # New containing block, use a new absolute list
         absolute_boxes = []
 
+    if gap == 'normal':
+        # 1em because in column context
+        gap = style['font_size']
+
     box = box.copy_with_children(box.children)
     box.position_y += collapse_margin(adjoining_margins) - box.margin_top
 
@@ -49,7 +53,8 @@ def columns_layout(context, box, bottom_space, skip_stack, containing_block,
         count = max(1, int(floor((box.width + gap) / (width + gap))))
         width = (box.width + gap) / count - gap
     else:  # overconstrained, with width != 'auto' and count != 'auto'
-        count = min(count, int(floor((box.width + gap) / (width + gap))))
+        count = max(
+            1, min(count, int(floor((box.width + gap) / (width + gap)))))
         width = (box.width + gap) / count - gap
 
     # Handle column-span property with the following structure:
