@@ -21,12 +21,13 @@ def float_layout(context, box, containing_block, absolute_boxes, fixed_boxes,
     """Set the width and position of floating ``box``."""
     from .block import block_container_layout
     from .flex import flex_layout
+    from .grid import grid_layout
 
     cb_width, cb_height = (containing_block.width, containing_block.height)
     resolve_percentages(box, (cb_width, cb_height))
 
     # TODO: This is only handled later in blocks.block_container_layout
-    # http://www.w3.org/TR/CSS21/visudet.html#normal-block
+    # https://www.w3.org/TR/CSS21/visudet.html#normal-block
     if cb_height == 'auto':
         cb_height = (
             containing_block.position_y - containing_block.content_box_y())
@@ -68,6 +69,12 @@ def float_layout(context, box, containing_block, absolute_boxes, fixed_boxes,
             skip_stack=skip_stack, containing_block=containing_block,
             page_is_empty=True, absolute_boxes=absolute_boxes,
             fixed_boxes=fixed_boxes)
+    elif isinstance(box, boxes.GridContainerBox):
+        box, resume_at, _, _, _ = grid_layout(
+            context, box, bottom_space=bottom_space,
+            skip_stack=skip_stack, containing_block=containing_block,
+            page_is_empty=True, absolute_boxes=absolute_boxes,
+            fixed_boxes=fixed_boxes)
     else:
         assert isinstance(box, boxes.BlockReplacedBox)
         resume_at = None
@@ -81,7 +88,7 @@ def float_layout(context, box, containing_block, absolute_boxes, fixed_boxes,
 
 def find_float_position(context, box, containing_block):
     """Get the right position of the float ``box``."""
-    # See http://www.w3.org/TR/CSS2/visuren.html#float-position
+    # See https://www.w3.org/TR/CSS2/visuren.html#float-position
 
     # Point 4 is already handled as box.position_y is set according to the
     # containing box top position, with collapsing margins handled
