@@ -1,7 +1,6 @@
 """PDF stream."""
 
 import io
-from functools import lru_cache
 from hashlib import md5
 
 import pydyf
@@ -13,19 +12,15 @@ from ..logger import LOGGER
 from ..matrix import Matrix
 from ..text.constants import PANGO_STRETCH_PERCENT
 from ..text.ffi import ffi, harfbuzz, pango, units_to_double
+from ..text.fonts import get_hb_object_data, get_pango_font_hb_face, get_pango_font_key
 
 
 class Font:
     def __init__(self, pango_font):
-        hb_font = pango.pango_font_get_hb_font(pango_font)
-        hb_face = harfbuzz.hb_font_get_face(hb_font)
-        hb_blob = ffi.gc(
-            harfbuzz.hb_face_reference_blob(hb_face),
-            harfbuzz.hb_blob_destroy)
-        with ffi.new('unsigned int *') as length:
-            hb_data = harfbuzz.hb_blob_get_data(hb_blob, length)
-            self.file_content = ffi.unpack(hb_data, int(length[0]))
-        self.index = harfbuzz.hb_face_get_index(hb_face)
+        self.hb_font = pango.pango_font_get_hb_font(pango_font)
+        self.hb_face = get_pango_font_hb_face(pango_font)
+        self.file_content = get_hb_object_data(self.hb_face)
+        self.index = harfbuzz.hb_face_get_index(self.hb_face)
 
         pango_metrics = pango.pango_font_get_metrics(pango_font, ffi.NULL)
         self.description = description = ffi.gc(
@@ -63,7 +58,7 @@ class Font:
         # Never use the built-in hash function here: it’s not stable
         self.hash = ''.join(
             chr(65 + letter % 26) for letter
-            in md5(description_string).digest()[:6])
+            in md5(description_string, usedforsecurity=False).digest()[:6])
 
         # Name
         fields = description_string.split(b' ')
@@ -96,16 +91,14 @@ class Font:
             self.bitmap = False
         else:
             self.bitmap = (
-                'EBDT' in self.ttfont and
-                'EBLC' in self.ttfont and (
-                    'glyf' not in self.ttfont or
-                    not self.ttfont['glyf'].glyphs))
+                'EBDT' in self.ttfont and 'EBLC' in self.ttfont and (
+                    'glyf' not in self.ttfont or not self.ttfont['glyf'].glyphs))
 
         # Various properties
         self.italic_angle = 0  # TODO: this should be different
-        self.upem = harfbuzz.hb_face_get_upem(hb_face)
-        self.png = harfbuzz.hb_ot_color_has_png(hb_face)
-        self.svg = harfbuzz.hb_ot_color_has_svg(hb_face)
+        self.upem = harfbuzz.hb_face_get_upem(self.hb_face)
+        self.png = harfbuzz.hb_ot_color_has_png(self.hb_face)
+        self.svg = harfbuzz.hb_ot_color_has_svg(self.hb_face)
         self.stemv = 80
         self.stemh = 80
         self.widths = {}
@@ -118,9 +111,6 @@ class Font:
             self.flags += 2 ** (7 - 1)  # Italic
         if b'Serif' in fields:
             self.flags += 2 ** (2 - 1)  # Serif
-        widths = self.widths.values()
-        if len(widths) > 1 and len(set(widths)) == 1:
-            self.flags += 2 ** (1 - 1)  # FixedPitch
 
     def clean(self, cmap, hinting):
         if self.ttfont is None:
@@ -333,15 +323,8 @@ class Stream(pydyf.Stream):
             'BM': f'/{mode}',
         }))
 
-    @lru_cache()
     def add_font(self, pango_font):
-        description = pango.pango_font_describe(pango_font)
-        mask = (
-            pango.PANGO_FONT_MASK_SIZE +
-            pango.PANGO_FONT_MASK_GRAVITY)
-        pango.pango_font_description_unset_fields(description, mask)
-        key = pango.pango_font_description_hash(description)
-        pango.pango_font_description_free(description)
+        key = get_pango_font_key(pango_font)
         if key not in self._fonts:
             self._fonts[key] = Font(pango_font)
         return self._fonts[key]
